/** 
 * popcount methods comparison
 * Author: KAlO2
 * Date: Mar 1, 2013
 */
 
#include <stdio.h>
#include <limits.h> /* for CHAR_BIT and UCHAR_MAX */
#include <time.h>

int iterated_popcnt(unsigned int n)
{
    int count=0;
    for(;n;n>>=1)
        count+=n&1u;
    return count;
}

int sparse_popcnt(unsigned int n)
{
    int count=0;
    while(n)
    {
        ++count;
        n&=(n-1);
    }
    return count;
}

int dense_popcnt(unsigned int n)
{
    int count=CHAR_BIT*sizeof(unsigned int);
    n^=(unsigned int)-1;
    while(n)
    {
        --count;
        n&=(n-1);
    }
    return count;
}

int lookup_popcnt(unsigned int n)
{
#define TBL_LEN (1<<CHAR_BIT)
#if 0

    static unsigned char TABLE[TBL_LEN]={0};
    for(int i=1;i<TBL_LEN;++i)
        TABLE[i]=TABLE[i>>1]+(i&1);
    
    unsigned char*p=(unsigned char*)&n;
    return TABLE[p[0]]+TABLE[p[1]]+TABLE[p[2]]+TABLE[p[3]];
    
#else 
# define CHAR_MASK ((unsigned char)(-1))
# define BIT2(n)      n,       n+1,       n+1,       n+2
# define BIT4(n) BIT2(n), BIT2(n+1), BIT2(n+1), BIT2(n+2)
# define BIT6(n) BIT4(n), BIT4(n+1), BIT4(n+1), BIT4(n+2)
# define BIT8(n) BIT6(n), BIT6(n+1), BIT6(n+1), BIT6(n+2)

    static const unsigned char TABLE[TBL_LEN]={BIT8(0)};
    
    return TABLE[ n                & UCHAR_MAX]+
           TABLE[(n>> CHAR_BIT)    & UCHAR_MAX]+
           TABLE[(n>>(CHAR_BIT*2)) & UCHAR_MAX]+
           TABLE[(n>>(CHAR_BIT*3)) & UCHAR_MAX];
           
#endif
#undef TBL_LEN
}

#define POW2(c) (1u<<(c))
#define MASK(c) (((unsigned int)(-1))/(POW2(POW2(c))+1u))
#define COUNT(x,c) ((x)&MASK(c)) + (((x)>>(POW2(c)))&MASK(c))

int parallel_popcnt(unsigned int n)
{
    n=COUNT(n,0);
    n=COUNT(n,1);
    n=COUNT(n,2);
    n=COUNT(n,3);
    n=COUNT(n,4);
/*  n=COUNT(n,5);  for 64-bit integers */
    return n ;
}

#define MASK_01010101 (((unsigned int)(-1))/3)
#define MASK_00110011 (((unsigned int)(-1))/5)
#define MASK_00001111 (((unsigned int)(-1))/17)

int nifty_popcnt(unsigned int n)
{
    n = (n & MASK_01010101) + ((n>>1) & MASK_01010101) ;
    n = (n & MASK_00110011) + ((n>>2) & MASK_00110011) ;
    n = (n & MASK_00001111) + ((n>>4) & MASK_00001111) ;        
    return n%255 ;
}

int hacker_popcnt(unsigned int n)
{
    n -= (n>>1) & 0x55555555;
    n  = (n & 0x33333333) + ((n>>2) & 0x33333333);
    n  = ((n>>4) + n) & 0x0F0F0F0F;
    n += n>>8;
    n += n>>16;
    return n & 0x0000003F;
}

int hakmem_popcnt(unsigned int n)
{
    unsigned int tmp;
    tmp = n - ((n>>1) & 033333333333) - ((n>>2) & 011111111111);
    return ((tmp + (tmp>>3)) & 030707070707) % 63;
}

int assembly_popcnt(unsigned int n)
{
#ifdef _MSC_VER /* Intel style assembly */
    __asm popcnt eax,n;
#elif __GNUC__ /* AT&T style assembly */
    __asm__("popcnt %0,%%eax"::"r"(n));
#else
# error "which assembly style does your compiler support, Intel or AT&T?"
#endif
}

int main(void)
{
    typedef int (*FUNC_BITCNT)(unsigned int);
    
    const struct Pair{
        FUNC_BITCNT pfunc;
        const char *name;
        }method[]=
    {
#define ELEMENT(n) {(n),#n}
        ELEMENT(iterated_popcnt),
        ELEMENT(sparse_popcnt),
        ELEMENT(dense_popcnt),
        ELEMENT(lookup_popcnt),
        ELEMENT(parallel_popcnt),
        ELEMENT(nifty_popcnt),
        ELEMENT(hacker_popcnt),
        ELEMENT(hakmem_popcnt),
        ELEMENT(assembly_popcnt)
#undef ELEMENT
    };

    time_t start,stop;
    for(int i=0;i<sizeof(method)/sizeof(method[0]);++i)
    {
        start=clock();
        for(unsigned int j=0;j<0xDEADD00D;++j)
            method[i].pfunc(j);
        stop=clock();
        double elapsed_time=(double)(stop-start)/CLOCKS_PER_SEC;
        
        printf("%d. method %15s uses %lfs\n", i, method[i].name, elapsed_time);
    }
}
